/*
,
  THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT

  LAWS OF THE UNITED STATES.  FOR LICENSING INFORMATION CONTACT:



  Christian Bischof or Lucas Roh, Mathematics and Computer Science Division,

  Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL 60439,

  {bischof,roh}@mcs.anl.gov.

*/

#if !defined(AD_GRAD_DYN_H)
#define AD_GRAD_DYN_H

/*
extern "C" {
#include "../include/run-alloc.h"
}
*/

#if !defined(AD_DERIV_H)
typedef struct {
        double value;
        double  *grad;
} DERIV_TYPE;
#define DERIV_val(a) ((a).value)
#define DERIV_grad(a) ((a).grad)
extern int ad_grad_size;
#endif

#define AD_PARAM_INIT(x) 

/* Use the following 4 macros in place of the run-time memory management */
/* For testing purposes only */
#define ad_adic_deriv_init(dsize,bsize)
#define ad_adic_deriv_final()
#define ad_adic_deriv_alloc() malloc(ad_grad_size*sizeof(double))
#define ad_adic_deriv_free(x) if (*x) free(*x)


#if defined(__cplusplus)
extern "C" {
#endif

/* ppx is a pointer to DERIV_TYPE */
#define VALIDATE(ppx) \
    if (!DERIV_grad(*ppx)) { \
        DERIV_grad(*ppx) = (double *)ad_adic_deriv_alloc(); \
    }

#define INVALIDATE(ppx) \
    if (DERIV_grad(*ppx)) { \
        ad_adic_deriv_free(&DERIV_grad(*ppx)); \
        DERIV_grad(*ppx) = (double*)0; \
    } 

#define INITIALIZE(ppx) \
    DERIV_grad(*ppx) = (double*)0;

#define IS_ZERO(ppx) \
    !(DERIV_grad(*ppx))

#define SET_ZERO_FLAG(flag, ppx, pos)\
    if (IS_ZERO(ppx)) {\
        flag |= (1<<pos);\
    }

#define DAXPY1(_ppz, _ca, _ppa) \
{\
  int _adic_i; \
  VALIDATE(_ppz); \
  for (_adic_i = 0; _adic_i < ad_grad_size; _adic_i++) {\
    DERIV_grad(*ppz)[_adic_i] = _ca * DERIV_grad(*_ppa)[_adic_i];\
  }\
}

#define DAXPY2(_ppz, _ca, _ppa, _cb, _ppb) \
{\
  int _adic_i; \
  VALIDATE(_ppz);\
  for (_adic_i = 0; _adic_i < ad_grad_size; _adic_i++) {\
    DERIV_grad(*ppz)[_adic_i] = \
       _ca * DERIV_grad(*_ppa)[_adic_i] + _cb * DERIV_grad(*_ppb)[_adic_i];\
  }\
}

#define DAXPY3(_ppz, _ca, _ppa, _cb, _ppb, _cc, _ppc) \
{\
  int _adic_i; \
  VALIDATE(_ppz);\
  for (_adic_i = 0; _adic_i < ad_grad_size; _adic_i++) {\
    DERIV_grad(*ppz)[_adic_i] = \
       _ca * DERIV_grad(*_ppa)[_adic_i] + _cb * DERIV_grad(*_ppb)[_adic_i] \
     + _cc * DERIV_grad(*_ppc)[_adic_i];\
  }\
}

#define DAXPY4(_ppz, _ca, _ppa, _cb, _ppb, _cc, _ppc, _cd, _ppd) \
{\
  int _adic_i; \
  VALIDATE(_ppz);\
  for (_adic_i = 0; _adic_i < ad_grad_size; _adic_i++) {\
    DERIV_grad(*ppz)[_adic_i] = \
       _ca * DERIV_grad(*_ppa)[_adic_i] + _cb * DERIV_grad(*_ppb)[_adic_i] \
     + _cc * DERIV_grad(*_ppc)[_adic_i] + _cd * DERIV_grad(*_ppd)[_adic_i];\
  }\
}

#define DAXPY5(_ppz, _ca, _ppa, _cb, _ppb, _cc, _ppc, _cd, _ppd, _ce, _ppe) \
{\
  int _adic_i; \
  VALIDATE(_ppz);\
  for (_adic_i = 0; _adic_i < ad_grad_size; _adic_i++) {\
    DERIV_grad(*ppz)[_adic_i] = \
       _ca * DERIV_grad(*_ppa)[_adic_i] + _cb * DERIV_grad(*_ppb)[_adic_i] \
     + _cc * DERIV_grad(*_ppc)[_adic_i] + _cd * DERIV_grad(*_ppd)[_adic_i] \
     + _ce * DERIV_grad(*_ppe)[_adic_i];\
  }\
}

void ad_grad_daxpy_init();
void ad_grad_daxpy_final();
#define ad_grad_daxpy_free(pz) ad_adic_deriv_free(pz)
void ad_grad_daxpy_0(DERIV_TYPE* ppz);

void ad_grad_daxpy_copy(DERIV_TYPE* ppz, DERIV_TYPE* pa);

void ad_grad_daxpy_1(DERIV_TYPE* pz, double a, DERIV_TYPE* pa);

void ad_grad_daxpy_2(DERIV_TYPE* ppz, double a, DERIV_TYPE* pa, 
                     double b, DERIV_TYPE* pb);

void ad_grad_daxpy_3(DERIV_TYPE* ppz, double a, DERIV_TYPE* pa, 
                     double b, DERIV_TYPE* pb, double c, DERIV_TYPE* pc);

void ad_grad_daxpy_n(int n, DERIV_TYPE* ppz, ...);
void ad_grad_daxpy_4(DERIV_TYPE* ppz, double ca, DERIV_TYPE* pa, double cb, DERIV_TYPE* pb, double cc, DERIV_TYPE* pc, double cd, DERIV_TYPE* pd);
void ad_grad_daxpy_5(DERIV_TYPE* ppz, double ca, DERIV_TYPE* pa, double cb, DERIV_TYPE* pb, double cc, DERIV_TYPE* pc, double cd, DERIV_TYPE* pd, double ce, DERIV_TYPE* pe);

#if defined(__cplusplus)
}
#endif
#endif /*AD_GRAD_DYN_H*/
